@@ -48,16 +48,19 @@
 #include "gdm-session-direct.h"
 #include "gdm-session.h"
 #include "gdm-session-private.h"
+#include "gdm-session-direct-glue.h"
 
 #include "gdm-session-record.h"
 #include "gdm-session-worker-job.h"
 
-#include "ck-connector.h"
-
 #define GDM_SESSION_DBUS_PATH         "/org/gnome/DisplayManager/Session"
 #define GDM_SESSION_DBUS_INTERFACE    "org.gnome.DisplayManager.Session"
 #define GDM_SESSION_DBUS_ERROR_CANCEL "org.gnome.DisplayManager.Session.Error.Cancel"
 
+#ifndef GDM_SESSION_DEFAULT_PATH
+#define GDM_SESSION_DEFAULT_PATH "/usr/local/bin:/usr/bin:/bin"
+#endif
+
 struct _GdmSessionDirectPrivate
 {
         /* per open scope */
@@ -65,12 +68,13 @@ struct _GdmSessionDirectPrivate
         char                *saved_session;
         char                *selected_language;
         char                *saved_language;
+        char                *selected_layout;
+        char                *saved_layout;
         char                *selected_user;
         char                *user_x11_authority_file;
 
         DBusMessage         *message_pending_reply;
         DBusConnection      *worker_connection;
-        CkConnector         *ckc;
 
         GdmSessionWorkerJob *job;
         GPid                 session_pid;
@@ -78,6 +82,8 @@ struct _GdmSessionDirectPrivate
         guint32              is_running : 1;
 
         /* object lifetime scope */
+        char                *id;
+        char                *display_id;
         char                *display_name;
         char                *display_hostname;
         char                *display_device;
@@ -87,10 +93,12 @@ struct _GdmSessionDirectPrivate
         DBusServer          *server;
         char                *server_address;
         GHashTable          *environment;
+        DBusGConnection     *connection;
 };
 
 enum {
         PROP_0,
+        PROP_DISPLAY_ID,
         PROP_DISPLAY_NAME,
         PROP_DISPLAY_HOSTNAME,
         PROP_DISPLAY_IS_LOCAL,
@@ -448,6 +456,135 @@ gdm_session_direct_handle_accreditation_failed (GdmSessionDirect *session,
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
+static const char **
+get_system_session_dirs (void)
+{
+        static const char *search_dirs[] = {
+                "/etc/X11/sessions/",
+                DMCONFDIR "/Sessions/",
+                DATADIR "/xsessions/",
+                DATADIR "/gdm/BuiltInSessions/",
+                NULL
+        };
+
+        return search_dirs;
+}
+
+static gboolean
+is_prog_in_path (const char *prog)
+{
+        char    *f;
+        gboolean ret;
+
+        f = g_find_program_in_path (prog);
+        ret = (f != NULL);
+        g_free (f);
+        return ret;
+}
+
+static gboolean
+get_session_command_for_file (const char *file,
+                              char      **command)
+{
+        GKeyFile   *key_file;
+        GError     *error;
+        char       *full_path;
+        char       *exec;
+        gboolean    ret;
+        gboolean    res;
+
+        exec = NULL;
+        ret = FALSE;
+        if (command != NULL) {
+                *command = NULL;
+        }
+
+        key_file = g_key_file_new ();
+
+        g_debug ("GdmSessionDirect: looking for session file '%s'", file);
+
+        error = NULL;
+        full_path = NULL;
+        res = g_key_file_load_from_dirs (key_file,
+                                         file,
+                                         get_system_session_dirs (),
+                                         &full_path,
+                                         G_KEY_FILE_NONE,
+                                         &error);
+        if (! res) {
+                g_debug ("GdmSessionDirect: File '%s' not found: %s", file, error->message);
+                g_error_free (error);
+                if (command != NULL) {
+                        *command = NULL;
+                }
+                goto out;
+        }
+
+        error = NULL;
+        res = g_key_file_get_boolean (key_file,
+                                      G_KEY_FILE_DESKTOP_GROUP,
+                                      G_KEY_FILE_DESKTOP_KEY_HIDDEN,
+                                      &error);
+        if (error == NULL && res) {
+                g_debug ("GdmSessionDirect: Session %s is marked as hidden", file);
+                goto out;
+        }
+
+        exec = g_key_file_get_string (key_file,
+                                      G_KEY_FILE_DESKTOP_GROUP,
+                                      G_KEY_FILE_DESKTOP_KEY_TRY_EXEC,
+                                      NULL);
+        if (exec != NULL) {
+                res = is_prog_in_path (exec);
+                g_free (exec);
+
+                if (! res) {
+                        g_debug ("GdmSessionDirect: Command not found: %s",
+                                 G_KEY_FILE_DESKTOP_KEY_TRY_EXEC);
+                        goto out;
+                }
+        }
+
+        error = NULL;
+        exec = g_key_file_get_string (key_file,
+                                      G_KEY_FILE_DESKTOP_GROUP,
+                                      G_KEY_FILE_DESKTOP_KEY_EXEC,
+                                      &error);
+        if (error != NULL) {
+                g_debug ("GdmSessionDirect: %s key not found: %s",
+                         G_KEY_FILE_DESKTOP_KEY_EXEC,
+                         error->message);
+                g_error_free (error);
+                goto out;
+        }
+
+        if (command != NULL) {
+                *command = g_strdup (exec);
+        }
+        ret = TRUE;
+
+out:
+        g_free (exec);
+
+        return ret;
+}
+
+static gboolean
+get_session_command_for_name (const char *name,
+                              char      **command)
+{
+        gboolean res;
+        char    *filename;
+
+        filename = g_strdup_printf ("%s.desktop", name);
+
+        command = NULL;
+        res = get_session_command_for_file (filename, command);
+        g_free (filename);
+
+        return res;
+}
+
 static const char *
 get_default_language_name (GdmSessionDirect *session)
 {
@@ -458,6 +595,66 @@ get_default_language_name (GdmSessionDirect *session)
     return setlocale (LC_MESSAGES, NULL);
 }
 
+static const char *
+get_default_layout_name (GdmSessionDirect *session)
+{
+    if (session->priv->saved_layout != NULL) {
+                return session->priv->saved_layout;
+    }
+
+    return "us";
+}
+
+static char *
+get_fallback_session_name (void)
+{
+        const char  **search_dirs;
+        int           i;
+        char         *name;
+
+        name = g_strdup ("gnome");
+        if (get_session_command_for_name (name, NULL)) {
+                return name;
+        }
+
+        search_dirs = get_system_session_dirs ();
+        for (i = 0; search_dirs[i] != NULL; i++) {
+                GDir *dir;
+                const char *base_name;
+
+                dir = g_dir_open (search_dirs[i], 0, NULL);
+
+                if (dir == NULL) {
+                        continue;
+                }
+
+                do {
+                        base_name = g_dir_read_name (dir);
+
+                        if (base_name == NULL) {
+                                break;
+                        }
+
+                        if (!g_str_has_suffix (base_name, ".desktop")) {
+                                continue;
+                        }
+
+                        if (get_session_command_for_file (base_name, NULL)) {
+                                g_free (name);
+
+                                name = g_strndup (base_name,
+                                                  strlen (base_name) -
+                                                  strlen (".desktop"));
+                                break;
+                        }
+                } while (base_name != NULL);
+
+                g_dir_close (dir);
+        }
+
+        return name;
+}
+
 static const char *
 get_default_session_name (GdmSessionDirect *session)
 {
@@ -465,7 +662,7 @@ get_default_session_name (GdmSessionDirect *session)
                 return session->priv->saved_session;
         }
 
-        return "gnome";
+        return get_fallback_session_name ();
 }
 
 static void
@@ -473,6 +670,8 @@ gdm_session_direct_defaults_changed (GdmSessionDirect *session)
 {
         _gdm_session_default_language_name_changed (GDM_SESSION (session),
                                                     get_default_language_name (session));
+        _gdm_session_default_layout_name_changed (GDM_SESSION (session),
+                                                  get_default_layout_name (session));
         _gdm_session_default_session_name_changed (GDM_SESSION (session),
                                                    get_default_session_name (session));
 }
@@ -493,6 +692,9 @@ gdm_session_direct_select_user (GdmSession *session,
 
         g_free (impl->priv->saved_language);
         impl->priv->saved_language = NULL;
+
+        g_free (impl->priv->saved_layout);
+        impl->priv->saved_layout = NULL;
 }
 
 static DBusHandlerResult
@@ -703,6 +905,10 @@ gdm_session_direct_handle_session_started (GdmSessionDirect *session,
         DBusError    error;
         int          pid;
 
+        pid = 0;
+
+        g_debug ("GdmSessionDirect: Handling SessionStarted");
+
         dbus_error_init (&error);
         if (! dbus_message_get_args (message, &error,
                                      DBUS_TYPE_INT32, &pid,
@@ -720,7 +926,7 @@ gdm_session_direct_handle_session_started (GdmSessionDirect *session,
         session->priv->session_pid = pid;
         session->priv->is_running = TRUE;
 
-        _gdm_session_session_started (GDM_SESSION (session));
+        _gdm_session_session_started (GDM_SESSION (session), pid);
 
         return DBUS_HANDLER_RESULT_HANDLED;
 }
@@ -842,6 +1048,39 @@ gdm_session_direct_handle_saved_language_name_read (GdmSessionDirect *session,
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
+static DBusHandlerResult
+gdm_session_direct_handle_saved_layout_name_read (GdmSessionDirect *session,
+                                                  DBusConnection   *connection,
+                                                  DBusMessage      *message)
+{
+        DBusMessage *reply;
+        DBusError    error;
+        const char  *layout_name;
+
+        dbus_error_init (&error);
+        if (! dbus_message_get_args (message, &error,
+                                     DBUS_TYPE_STRING, &layout_name,
+                                     DBUS_TYPE_INVALID)) {
+                g_warning ("ERROR: %s", error.message);
+        }
+
+        reply = dbus_message_new_method_return (message);
+        dbus_connection_send (connection, reply, NULL);
+        dbus_message_unref (reply);
+
+        if (strcmp (layout_name,
+                    get_default_layout_name (session)) != 0) {
+                g_free (session->priv->saved_layout);
+                session->priv->saved_layout = g_strdup (layout_name);
+
+                _gdm_session_default_layout_name_changed (GDM_SESSION (session),
+                                                          layout_name);
+        }
+
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
+
 static DBusHandlerResult
 gdm_session_direct_handle_saved_session_name_read (GdmSessionDirect *session,
                                                    DBusConnection   *connection,
@@ -862,6 +1101,14 @@ gdm_session_direct_handle_saved_session_name_read (GdmSessionDirect *session,
         dbus_connection_send (connection, reply, NULL);
         dbus_message_unref (reply);
 
+        if (! get_session_command_for_name (session_name, NULL)) {
+                /* ignore sessions that don't exist */
+                g_debug ("GdmSessionDirect: not using invalid .dmrc session: %s", session_name);
+                g_free (session->priv->saved_session);
+                session->priv->saved_session = NULL;
+                goto out;
+        }
+
         if (strcmp (session_name,
                     get_default_session_name (session)) != 0) {
                 g_free (session->priv->saved_session);
@@ -870,7 +1117,7 @@ gdm_session_direct_handle_saved_session_name_read (GdmSessionDirect *session,
                 _gdm_session_default_session_name_changed (GDM_SESSION (session),
                                                            session_name);
         }
-
+ out:
         return DBUS_HANDLER_RESULT_HANDLED;
 }
 
@@ -911,6 +1158,18 @@ session_worker_message (DBusConnection *connection,
                 return gdm_session_direct_handle_accredited (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "AccreditationFailed")) {
                 return gdm_session_direct_handle_accreditation_failed (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "Authenticated")) {
+                return gdm_session_direct_handle_authenticated (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "AuthenticationFailed")) {
+                return gdm_session_direct_handle_authentication_failed (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "Authorized")) {
+                return gdm_session_direct_handle_authorized (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "AuthorizationFailed")) {
+                return gdm_session_direct_handle_authorization_failed (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "Accredited")) {
+                return gdm_session_direct_handle_accredited (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "AccreditationFailed")) {
+                return gdm_session_direct_handle_accreditation_failed (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "UsernameChanged")) {
                 return gdm_session_direct_handle_username_changed (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SessionStarted")) {
@@ -923,6 +1182,8 @@ session_worker_message (DBusConnection *connection,
                 return gdm_session_direct_handle_session_died (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedLanguageNameRead")) {
                 return gdm_session_direct_handle_saved_language_name_read (session, connection, message);
+        } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedLayoutNameRead")) {
+                return gdm_session_direct_handle_saved_layout_name_read (session, connection, message);
         } else if (dbus_message_is_method_call (message, GDM_SESSION_DBUS_INTERFACE, "SavedSessionNameRead")) {
                 return gdm_session_direct_handle_saved_session_name_read (session, connection, message);
         }
@@ -1002,6 +1263,7 @@ do_introspect (DBusConnection *connection,
                                "    </method>\n"
                                "    <method name=\"SessionStarted\">\n"
                                "      <arg name=\"pid\" direction=\"in\" type=\"i\"/>\n"
+                               "      <arg name=\"environment\" direction=\"in\" type=\"as\"/>\n"
                                "    </method>\n"
                                "    <method name=\"SessionExited\">\n"
                                "      <arg name=\"code\" direction=\"in\" type=\"i\"/>\n"
@@ -1032,12 +1294,18 @@ do_introspect (DBusConnection *connection,
                                "    </signal>\n"
                                "    <signal name=\"EstablishCredentials\">\n"
                                "    </signal>\n"
-                               "    <signal name=\"RenewCredentials\">\n"
+                               "    <signal name=\"RefreshCredentials\">\n"
                                "    </signal>\n"
                                "    <signal name=\"SetEnvironmentVariable\">\n"
                                "      <arg name=\"name\" type=\"s\"/>\n"
                                "      <arg name=\"value\" type=\"s\"/>\n"
                                "    </signal>\n"
+                               "    <signal name=\"SetLanguageName\">\n"
+                               "      <arg name=\"language_name\" type=\"s\"/>\n"
+                               "    </signal>\n"
+                               "    <signal name=\"SetSessionName\">\n"
+                               "      <arg name=\"session_name\" type=\"s\"/>\n"
+                               "    </signal>\n"
                                "    <signal name=\"StartProgram\">\n"
                                "      <arg name=\"command\" type=\"s\"/>\n"
                                "    </signal>\n"
@@ -1276,6 +1544,7 @@ gdm_session_direct_init (GdmSessionDirect *session)
                                                             (GDestroyNotify) g_free);
 
         setup_server (session);
+
 }
 
 static void
@@ -1563,8 +1832,8 @@ gdm_session_direct_accredit (GdmSession *session,
         case GDM_SESSION_CRED_ESTABLISH:
                 send_dbus_void_signal (impl, "EstablishCredentials");
                 break;
-        case GDM_SESSION_CRED_RENEW:
-                send_dbus_void_signal (impl, "RenewCredentials");
+        case GDM_SESSION_CRED_REFRESH:
+                send_dbus_void_signal (impl, "RefreshCredentials");
                 break;
         default:
                 g_assert_not_reached ();
@@ -1603,115 +1872,6 @@ send_environment (GdmSessionDirect *session)
                               session);
 }
 
-static gboolean
-is_prog_in_path (const char *prog)
-{
-        char    *f;
-        gboolean ret;
-
-        f = g_find_program_in_path (prog);
-        ret = (f != NULL);
-        g_free (f);
-        return ret;
-}
-
-static gboolean
-get_session_command_for_file (const char *file,
-                              char      **command)
-{
-        GKeyFile   *key_file;
-        GError     *error;
-        char       *full_path;
-        char       *exec;
-        gboolean    ret;
-        gboolean    res;
-        const char *search_dirs[] = {
-                "/etc/X11/sessions/",
-                DMCONFDIR "/Sessions/",
-                DATADIR "/gdm/BuiltInSessions/",
-                DATADIR "/xsessions/",
-                NULL
-        };
-
-        exec = NULL;
-        ret = FALSE;
-        if (command != NULL) {
-                *command = NULL;
-        }
-
-        key_file = g_key_file_new ();
-
-        g_debug ("GdmSessionDirect: looking for session file '%s'", file);
-
-        error = NULL;
-        full_path = NULL;
-        res = g_key_file_load_from_dirs (key_file,
-                                         file,
-                                         search_dirs,
-                                         &full_path,
-                                         G_KEY_FILE_NONE,
-                                         &error);
-        if (! res) {
-                g_debug ("GdmSessionDirect: File '%s' not found: %s", file, error->message);
-                g_error_free (error);
-                if (command != NULL) {
-                        *command = NULL;
-                }
-                goto out;
-        }
-
-        error = NULL;
-        res = g_key_file_get_boolean (key_file,
-                                      G_KEY_FILE_DESKTOP_GROUP,
-                                      G_KEY_FILE_DESKTOP_KEY_HIDDEN,
-                                      &error);
-        if (error == NULL && res) {
-                g_debug ("GdmSessionDirect: Session %s is marked as hidden", file);
-                goto out;
-        }
-
-        error = NULL;
-        exec = g_key_file_get_string (key_file,
-                                      G_KEY_FILE_DESKTOP_GROUP,
-                                      G_KEY_FILE_DESKTOP_KEY_TRY_EXEC,
-                                      &error);
-        if (exec == NULL) {
-                g_debug ("GdmSessionDirect: %s key not found", G_KEY_FILE_DESKTOP_KEY_TRY_EXEC);
-                goto out;
-        }
-
-        res = is_prog_in_path (exec);
-        g_free (exec);
-
-        if (! res) {
-                g_debug ("GdmSessionDirect: Command not found: %s", G_KEY_FILE_DESKTOP_KEY_TRY_EXEC);
-                goto out;
-        }
-
-        error = NULL;
-        exec = g_key_file_get_string (key_file,
-                                      G_KEY_FILE_DESKTOP_GROUP,
-                                      G_KEY_FILE_DESKTOP_KEY_EXEC,
-                                      &error);
-        if (error != NULL) {
-                g_debug ("GdmSessionDirect: %s key not found: %s",
-                         G_KEY_FILE_DESKTOP_KEY_EXEC,
-                         error->message);
-                g_error_free (error);
-                goto out;
-        }
-
-        if (command != NULL) {
-                *command = g_strdup (exec);
-        }
-        ret = TRUE;
-
-out:
-        g_free (exec);
-
-        return ret;
-}
-
 static const char *
 get_language_name (GdmSessionDirect *session)
 {
@@ -1722,9 +1882,21 @@ get_language_name (GdmSessionDirect *session)
         return get_default_language_name (session);
 }
 
+static const char *
+get_layout_name (GdmSessionDirect *session)
+{
+        if (session->priv->selected_layout != NULL) {
+                return session->priv->selected_layout;
+        }
+
+        return get_default_layout_name (session);
+}
+
 static const char *
 get_session_name (GdmSessionDirect *session)
 {
+        /* FIXME: test the session names before we use them? */
+
         if (session->priv->selected_session != NULL) {
                 return session->priv->selected_session;
         }
@@ -1753,73 +1925,6 @@ get_session_command (GdmSessionDirect *session)
         return command;
 }
 
-static gboolean
-open_ck_session (GdmSessionDirect *session)
-{
-        struct passwd *pwent;
-        gboolean       ret;
-        int            res;
-        DBusError      error;
-        const char     *display_name;
-        const char     *display_device;
-        const char     *display_hostname;
-
-        if (session->priv->display_name != NULL) {
-                display_name = session->priv->display_name;
-        } else {
-                display_name = "";
-        }
-        if (session->priv->display_hostname != NULL) {
-                display_hostname = session->priv->display_hostname;
-        } else {
-                display_hostname = "";
-        }
-        if (session->priv->display_device != NULL) {
-                display_device = session->priv->display_device;
-        } else {
-                display_device = "";
-        }
-
-        g_assert (session->priv->selected_user != NULL);
-
-        pwent = getpwnam (session->priv->selected_user);
-        if (pwent == NULL) {
-                return FALSE;
-        }
-
-        session->priv->ckc = ck_connector_new ();
-        if (session->priv->ckc == NULL) {
-                g_warning ("Couldn't create new ConsoleKit connector");
-                goto out;
-        }
-
-        dbus_error_init (&error);
-        res = ck_connector_open_session_with_parameters (session->priv->ckc,
-                                                         &error,
-                                                         "unix-user", &pwent->pw_uid,
-                                                         "x11-display", &display_name,
-                                                         "x11-display-device", &display_device,
-                                                         "remote-host-name", &display_hostname,
-                                                         "is-local", &session->priv->display_is_local,
-                                                         NULL);
-
-        if (! res) {
-                if (dbus_error_is_set (&error)) {
-                        g_warning ("%s\n", error.message);
-                        dbus_error_free (&error);
-                } else {
-                        g_warning ("cannot open CK session: OOM, D-Bus system bus not available,\n"
-                                   "ConsoleKit not available or insufficient privileges.\n");
-                }
-                goto out;
-        }
-
-        ret = TRUE;
-
- out:
-        return ret;
-}
-
 static void
 gdm_session_direct_set_environment_variable (GdmSessionDirect *session,
                                              const char       *key,
@@ -1838,15 +1943,6 @@ gdm_session_direct_set_environment_variable (GdmSessionDirect *session,
 static void
 setup_session_environment (GdmSessionDirect *session)
 {
-        const char *session_cookie;
-        gboolean    res;
-
-        session_cookie = NULL;
-        res = open_ck_session (session);
-        if (res) {
-                session_cookie = ck_connector_get_cookie (session->priv->ckc);
-        }
-
         gdm_session_direct_set_environment_variable (session,
                                                      "GDMSESSION",
                                                      get_session_name (session));
@@ -1861,14 +1957,13 @@ setup_session_environment (GdmSessionDirect *session)
                                                      "GDM_LANG",
                                                      get_language_name (session));
 
+        gdm_session_direct_set_environment_variable (session,
+                                                     "GDM_KEYBOARD_LAYOUT",
+                                                     get_layout_name (session));
+
         gdm_session_direct_set_environment_variable (session,
                                                      "DISPLAY",
                                                      session->priv->display_name);
-        if (session_cookie != NULL) {
-                gdm_session_direct_set_environment_variable (session,
-                                                             "XDG_SESSION_COOKIE",
-                                                             session_cookie);
-        }
 
         if (session->priv->user_x11_authority_file != NULL) {
                 gdm_session_direct_set_environment_variable (session,
@@ -1876,9 +1971,14 @@ setup_session_environment (GdmSessionDirect *session)
                                                              session->priv->user_x11_authority_file);
         }
 
+        /* FIXME: We do this here and in the session worker.  We should consolidate
+         * somehow.
+         */
         gdm_session_direct_set_environment_variable (session,
                                                      "PATH",
-                                                     "/bin:/usr/bin:" BINDIR);
+                                                     strcmp (BINDIR, "/usr/bin") == 0?
+                                                     GDM_SESSION_DEFAULT_PATH :
+                                                     BINDIR ":" GDM_SESSION_DEFAULT_PATH);
 
 }
 
@@ -1912,12 +2012,6 @@ gdm_session_direct_close (GdmSession *session)
 
         g_debug ("GdmSessionDirect: Closing session");
 
-        if (impl->priv->ckc != NULL) {
-                ck_connector_close_session (impl->priv->ckc, NULL);
-                ck_connector_unref (impl->priv->ckc);
-                impl->priv->ckc = NULL;
-        }
-
         if (impl->priv->job != NULL) {
                 if (impl->priv->is_running) {
                         gdm_session_record_logout (impl->priv->session_pid,
@@ -1945,6 +2039,12 @@ gdm_session_direct_close (GdmSession *session)
         g_free (impl->priv->saved_language);
         impl->priv->saved_language = NULL;
 
+        g_free (impl->priv->selected_layout);
+        impl->priv->selected_layout = NULL;
+
+        g_free (impl->priv->saved_layout);
+        impl->priv->saved_layout = NULL;
+
         g_free (impl->priv->user_x11_authority_file);
         impl->priv->user_x11_authority_file = NULL;
 
@@ -1997,6 +2097,9 @@ gdm_session_direct_select_session (GdmSession *session,
         } else {
                 impl->priv->selected_session = g_strdup (text);
         }
+
+        send_dbus_string_signal (impl, "SetSessionName",
+                                 get_session_name (impl));
 }
 
 static void
@@ -2012,6 +2115,35 @@ gdm_session_direct_select_language (GdmSession *session,
         } else {
                 impl->priv->selected_language = g_strdup (text);
         }
+
+        send_dbus_string_signal (impl, "SetLanguageName",
+                                 get_language_name (impl));
+}
+
+static void
+gdm_session_direct_select_layout (GdmSession *session,
+                                  const char *text)
+{
+        GdmSessionDirect *impl = GDM_SESSION_DIRECT (session);
+
+        g_free (impl->priv->selected_layout);
+
+        if (strcmp (text, "__previous") == 0) {
+                impl->priv->selected_layout = NULL;
+        } else {
+                impl->priv->selected_layout = g_strdup (text);
+        }
+
+        send_dbus_string_signal (impl, "SetLayoutName",
+                                 get_layout_name (impl));
+}
+
+static void
+_gdm_session_direct_set_display_id (GdmSessionDirect *session,
+                                    const char       *id)
+{
+        g_free (session->priv->display_id);
+        session->priv->display_id = g_strdup (id);
 }
 
 /* At some point we may want to read these right from
@@ -2075,6 +2207,9 @@ gdm_session_direct_set_property (GObject      *object,
         self = GDM_SESSION_DIRECT (object);
 
         switch (prop_id) {
+        case PROP_DISPLAY_ID:
+                _gdm_session_direct_set_display_id (self, g_value_get_string (value));
+                break;
         case PROP_DISPLAY_NAME:
                 _gdm_session_direct_set_display_name (self, g_value_get_string (value));
                 break;
@@ -2110,6 +2245,9 @@ gdm_session_direct_get_property (GObject    *object,
         self = GDM_SESSION_DIRECT (object);
 
         switch (prop_id) {
+        case PROP_DISPLAY_ID:
+                g_value_set_string (value, self->priv->display_id);
+                break;
         case PROP_DISPLAY_NAME:
                 g_value_set_string (value, self->priv->display_name);
                 break;
@@ -2145,6 +2283,9 @@ gdm_session_direct_dispose (GObject *object)
 
         gdm_session_direct_close (GDM_SESSION (session));
 
+        g_free (session->priv->display_id);
+        session->priv->display_id = NULL;
+
         g_free (session->priv->display_name);
         session->priv->display_name = NULL;
 
@@ -2182,11 +2323,15 @@ gdm_session_direct_finalize (GObject *object)
 
         session = GDM_SESSION_DIRECT (object);
 
+        g_free (session->priv->id);
+
         g_free (session->priv->selected_user);
         g_free (session->priv->selected_session);
         g_free (session->priv->saved_session);
         g_free (session->priv->selected_language);
         g_free (session->priv->saved_language);
+        g_free (session->priv->selected_layout);
+        g_free (session->priv->saved_layout);
 
         parent_class = G_OBJECT_CLASS (gdm_session_direct_parent_class);
 
@@ -2194,6 +2339,58 @@ gdm_session_direct_finalize (GObject *object)
                 parent_class->finalize (object);
 }
 
+static gboolean
+register_session (GdmSessionDirect *session)
+{
+        GError *error;
+
+        error = NULL;
+        session->priv->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
+        if (session->priv->connection == NULL) {
+                if (error != NULL) {
+                        g_critical ("error getting system bus: %s", error->message);
+                        g_error_free (error);
+                }
+                exit (1);
+        }
+
+        dbus_g_connection_register_g_object (session->priv->connection, session->priv->id, G_OBJECT (session));
+
+        return TRUE;
+}
+
+static GObject *
+gdm_session_direct_constructor (GType                  type,
+                                guint                  n_construct_properties,
+                                GObjectConstructParam *construct_properties)
+{
+        GdmSessionDirect *session;
+        gboolean          res;
+        const char       *id;
+
+        session = GDM_SESSION_DIRECT (G_OBJECT_CLASS (gdm_session_direct_parent_class)->constructor (type,
+                                                                                          n_construct_properties,
+                                                                                          construct_properties));
+
+        /* Always match the session id with the master */
+        id = NULL;
+        if (g_str_has_prefix (session->priv->display_id, "/org/gnome/DisplayManager/Display")) {
+                id = session->priv->display_id + strlen ("/org/gnome/DisplayManager/Display");
+        }
+
+        g_assert (id != NULL);
+
+        session->priv->id = g_strdup_printf ("/org/gnome/DisplayManager/Session%s", id);
+        g_debug ("GdmSessionDirect: Registering %s", session->priv->id);
+
+        res = register_session (session);
+        if (! res) {
+                g_warning ("Unable to register session with system bus");
+        }
+
+        return G_OBJECT (session);
+}
+
 static void
 gdm_session_iface_init (GdmSessionIface *iface)
 {
@@ -2210,6 +2407,7 @@ gdm_session_iface_init (GdmSessionIface *iface)
         iface->answer_query = gdm_session_direct_answer_query;
         iface->select_session = gdm_session_direct_select_session;
         iface->select_language = gdm_session_direct_select_language;
+        iface->select_layout = gdm_session_direct_select_layout;
         iface->select_user = gdm_session_direct_select_user;
 }
 
@@ -2222,11 +2420,19 @@ gdm_session_direct_class_init (GdmSessionDirectClass *session_class)
 
         object_class->get_property = gdm_session_direct_get_property;
         object_class->set_property = gdm_session_direct_set_property;
+        object_class->constructor = gdm_session_direct_constructor;
         object_class->dispose = gdm_session_direct_dispose;
         object_class->finalize = gdm_session_direct_finalize;
 
         g_type_class_add_private (session_class, sizeof (GdmSessionDirectPrivate));
 
+        g_object_class_install_property (object_class,
+                                         PROP_DISPLAY_ID,
+                                         g_param_spec_string ("display-id",
+                                                              "display id",
+                                                              "display id",
+                                                              NULL,
+                                                              G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
         g_object_class_install_property (object_class,
                                          PROP_DISPLAY_NAME,
                                          g_param_spec_string ("display-name",
@@ -2272,10 +2478,12 @@ gdm_session_direct_class_init (GdmSessionDirectClass *session_class)
                                                               G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
 
 
+        dbus_g_object_type_install_info (GDM_TYPE_SESSION_DIRECT, &dbus_glib_gdm_session_direct_object_info);
 }
 
 GdmSessionDirect *
-gdm_session_direct_new (const char *display_name,
+gdm_session_direct_new (const char *display_id,
+                        const char *display_name,
                         const char *display_hostname,
                         const char *display_device,
                         const char *display_x11_authority_file,
@@ -2284,6 +2492,7 @@ gdm_session_direct_new (const char *display_name,
         GdmSessionDirect *session;
 
         session = g_object_new (GDM_TYPE_SESSION_DIRECT,
+                                "display-id", display_id,
                                 "display-name", display_name,
                                 "display-hostname", display_hostname,
                                 "display-device", display_device,
@@ -2293,3 +2502,41 @@ gdm_session_direct_new (const char *display_name,
 
         return session;
 }
+
+gboolean
+gdm_session_direct_restart (GdmSessionDirect *session,
+                            GError          **error)
+{
+        gboolean ret;
+
+        ret = TRUE;
+        g_debug ("GdmSessionDirect: Request to restart session");
+
+        return ret;
+}
+
+gboolean
+gdm_session_direct_stop (GdmSessionDirect *session,
+                         GError          **error)
+{
+        gboolean ret;
+
+        ret = TRUE;
+
+        g_debug ("GdmSessionDirect: Request to stop session");
+
+        return ret;
+}
+
+gboolean
+gdm_session_direct_detach (GdmSessionDirect *session,
+                           GError          **error)
+{
+        gboolean ret;
+
+        ret = TRUE;
+
+        g_debug ("GdmSessionDirect: Request to detach session");
+
+        return ret;
+}
