@@ -99,8 +99,14 @@ on_session_started (GdmSession       *session,
                     int               pid,
                     GdmSimpleSlave   *slave)
 {
+        char *username;
+
         g_debug ("GdmSimpleSlave: session started %d", pid);
 
+        /* Run the PreSession script. gdmslave suspends until script has terminated */
+        username = gdm_session_direct_get_username (slave->priv->session);
+        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/PreSession", username);
+
         /* FIXME: should we do something here? */
 }
 
@@ -109,8 +115,14 @@ on_session_exited (GdmSession     *session,
                    int             exit_code,
                    GdmSimpleSlave *slave)
 {
+        char *username;
+
         g_debug ("GdmSimpleSlave: session exited with code %d\n", exit_code);
 
+        /* Run the PostSession script. gdmslave suspends until script has terminated */
+        username = gdm_session_direct_get_username (slave->priv->session);
+        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/PostSession", username);
+
         gdm_slave_stopped (GDM_SLAVE (slave));
 }
 
@@ -353,6 +365,9 @@ on_session_accredited (GdmSession     *session,
                        GdmSimpleSlave *slave)
 {
         queue_start_session (slave);
+
+        /* Run the PostLogin script. gdmslave suspends until script has terminated */
+        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/PostLogin", "gdm");
 }
 
 static void
@@ -484,6 +499,16 @@ on_default_language_name_changed (GdmSession     *session,
         gdm_greeter_server_default_language_name_changed (slave->priv->greeter_server, text);
 }
 
+static void
+on_default_layout_name_changed (GdmSession     *session,
+                                const char     *text,
+                                GdmSimpleSlave *slave)
+{
+        g_debug ("GdmSimpleSlave: Default layout name changed: %s", text);
+
+        gdm_greeter_server_default_layout_name_changed (slave->priv->greeter_server, text);
+}
+
 static void
 on_default_session_name_changed (GdmSession     *session,
                                  const char     *text,
@@ -618,6 +643,11 @@ create_new_session (GdmSimpleSlave *slave)
                           G_CALLBACK (on_default_language_name_changed),
                           slave);
 
+        g_signal_connect (slave->priv->session,
+                          "default-layout-name-changed",
+                          G_CALLBACK (on_default_layout_name_changed),
+                          slave);
+
         g_signal_connect (slave->priv->session,
                           "default-session-name-changed",
                           G_CALLBACK (on_default_session_name_changed),
@@ -712,6 +742,14 @@ on_greeter_language_selected (GdmGreeterServer *greeter_server,
         gdm_session_select_language (GDM_SESSION (slave->priv->session), text);
 }
 
+static void
+on_greeter_layout_selected (GdmGreeterServer *greeter_server,
+                            const char       *text,
+                            GdmSimpleSlave   *slave)
+{
+        gdm_session_select_layout (GDM_SESSION (slave->priv->session), text);
+}
+
 static void
 on_greeter_user_selected (GdmGreeterServer *greeter_server,
                           const char       *text,
@@ -817,7 +855,7 @@ run_greeter (GdmSimpleSlave *slave)
         }
 
         /* Run the init script. gdmslave suspends until script has terminated */
-        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/Init/Default", "gdm");
+        gdm_slave_run_script (GDM_SLAVE (slave), GDMCONFDIR "/Init", "gdm");
 
         create_new_session (slave);
 
@@ -846,6 +884,10 @@ run_greeter (GdmSimpleSlave *slave)
                           "language-selected",
                           G_CALLBACK (on_greeter_language_selected),
                           slave);
+        g_signal_connect (slave->priv->greeter_server,
+                          "layout-selected",
+                          G_CALLBACK (on_greeter_layout_selected),
+                          slave);
         g_signal_connect (slave->priv->greeter_server,
                           "user-selected",
                           G_CALLBACK (on_greeter_user_selected),
@@ -976,8 +1018,18 @@ gdm_simple_slave_run (GdmSimpleSlave *slave)
          * exist */
         if (display_is_local) {
                 gboolean res;
+                gboolean disable_tcp;
 
                 slave->priv->server = gdm_server_new (display_name, auth_file);
+
+                disable_tcp = TRUE;
+                if (gdm_settings_client_get_boolean (GDM_KEY_DISALLOW_TCP,
+                                                     &disable_tcp)) {
+                        g_object_set (slave->priv->server,
+                                      "disable-tcp", disable_tcp,
+                                      NULL);
+                }
+
                 g_signal_connect (slave->priv->server,
                                   "exited",
                                   G_CALLBACK (on_server_exited),
