@@ -35,6 +35,7 @@
 
 #include <glib.h>
 #include <glib/gi18n.h>
+#include <glib/gstdio.h>
 #include <glib-object.h>
 #define DBUS_API_SUBJECT_TO_CHANGE
 #include <dbus/dbus-glib.h>
@@ -316,6 +317,7 @@ get_welcome_environment (GdmWelcomeSession *welcome_session)
         g_hash_table_insert (hash, g_strdup ("PATH"), g_strdup (g_getenv ("PATH")));
 
         g_hash_table_insert (hash, g_strdup ("RUNNING_UNDER_GDM"), g_strdup ("true"));
+        g_hash_table_insert (hash, g_strdup ("GVFS_DISABLE_FUSE"), g_strdup ("1"));
 
         g_hash_table_foreach (hash, (GHFunc)listify_hash, env);
         g_hash_table_destroy (hash);
@@ -329,10 +331,16 @@ get_welcome_environment (GdmWelcomeSession *welcome_session)
 static gboolean
 stop_dbus_daemon (GdmWelcomeSession *welcome_session)
 {
+        int res;
+
         if (welcome_session->priv->dbus_pid > 0) {
                 g_debug ("GdmWelcomeSession: Stopping D-Bus daemon");
-                gdm_signal_pid (-1 * welcome_session->priv->dbus_pid, SIGTERM);
-                welcome_session->priv->dbus_pid = 0;
+                res = gdm_signal_pid (-1 * welcome_session->priv->dbus_pid, SIGTERM);
+                if (res < 0) {
+                        g_warning ("Unable to kill D-Bus daemon");
+                } else {
+                        welcome_session->priv->dbus_pid = 0;
+                }
         }
         return TRUE;
 }
@@ -371,6 +379,7 @@ welcome_session_child_watch (GPid               pid,
 typedef struct {
         const char *user_name;
         const char *group_name;
+        const char *log_file;
 } SpawnChildData;
 
 static void
@@ -437,12 +446,35 @@ spawn_child_setup (SpawnChildData *data)
                          (guint) getpid (), g_strerror (errno));
                 _exit (2);
         }
+
+        if (data->log_file != NULL) {
+                int logfd;
+
+                if (g_access (data->log_file, R_OK | W_OK) == 0) {
+                        char *filename_old;
+
+                        filename_old = g_strdup_printf ("%s.old", data->log_file);
+                        VE_IGNORE_EINTR (g_unlink (filename_old));
+                        VE_IGNORE_EINTR (g_rename (data->log_file, filename_old));
+                        g_free (filename_old);
+                }
+
+                VE_IGNORE_EINTR (g_unlink (data->log_file));
+                VE_IGNORE_EINTR (logfd = open (data->log_file, O_CREAT|O_TRUNC|O_WRONLY|O_EXCL, 0644));
+
+                if (logfd != -1) {
+                        VE_IGNORE_EINTR (dup2 (logfd, 1));
+                        VE_IGNORE_EINTR (dup2 (logfd, 2));
+                        close (logfd);
+                }
+        }
 }
 
 static gboolean
 spawn_command_line_sync_as_user (const char *command_line,
                                  const char *user_name,
                                  const char *group_name,
+                                 const char *log_file,
                                  char       **env,
                                  char       **std_output,
                                  char       **std_error,
@@ -467,6 +499,7 @@ spawn_command_line_sync_as_user (const char *command_line,
 
         data.user_name = user_name;
         data.group_name = group_name;
+        data.log_file = log_file;
 
         local_error = NULL;
         res = g_spawn_sync (NULL,
@@ -497,6 +530,7 @@ static gboolean
 spawn_command_line_async_as_user (const char *command_line,
                                   const char *user_name,
                                   const char *group_name,
+                                  const char *log_file,
                                   char      **env,
                                   GPid       *child_pid,
                                   GError    **error)
@@ -519,6 +553,7 @@ spawn_command_line_async_as_user (const char *command_line,
 
         data.user_name = user_name;
         data.group_name = group_name;
+        data.log_file = log_file;
 
         local_error = NULL;
         res = g_spawn_async (NULL,
@@ -529,6 +564,7 @@ spawn_command_line_async_as_user (const char *command_line,
                              &data,
                              child_pid,
                              &local_error);
+
         if (! res) {
                 g_warning ("Could not spawn command: %s", local_error->message);
                 g_propagate_error (error, local_error);
@@ -583,7 +619,7 @@ parse_dbus_launch_output (const char *output,
         error = NULL;
         re = g_regex_new ("DBUS_SESSION_BUS_ADDRESS=(.+)\nDBUS_SESSION_BUS_PID=([0-9]+)", 0, 0, &error);
         if (re == NULL) {
-                g_critical (error->message);
+                g_critical ("%s", error->message);
         }
 
         g_regex_match (re, output, 0, &match_info);
@@ -636,6 +672,7 @@ start_dbus_daemon (GdmWelcomeSession *welcome_session)
         res = spawn_command_line_sync_as_user (DBUS_LAUNCH_COMMAND,
                                                welcome_session->priv->user_name,
                                                welcome_session->priv->group_name,
+                                               NULL, /* log file */
                                                (char **)env->pdata,
                                                &std_out,
                                                &std_err,
@@ -670,6 +707,8 @@ gdm_welcome_session_spawn (GdmWelcomeSession *welcome_session)
         GPtrArray       *env;
         gboolean         ret;
         gboolean         res;
+        char            *log_path;
+        char            *log_file;
 
         ret = FALSE;
 
@@ -688,9 +727,14 @@ gdm_welcome_session_spawn (GdmWelcomeSession *welcome_session)
 
         error = NULL;
 
+        log_file = g_strdup_printf ("%s-greeter.log", welcome_session->priv->x11_display_name);
+        log_path = g_build_filename (LOGDIR, log_file, NULL);
+        g_free (log_file);
+
         ret = spawn_command_line_async_as_user (welcome_session->priv->command,
                                                 welcome_session->priv->user_name,
                                                 welcome_session->priv->group_name,
+                                                log_path,
                                                 (char **)env->pdata,
                                                 &welcome_session->priv->pid,
                                                 &error);
@@ -698,6 +742,8 @@ gdm_welcome_session_spawn (GdmWelcomeSession *welcome_session)
         g_ptr_array_foreach (env, (GFunc)g_free, NULL);
         g_ptr_array_free (env, TRUE);
 
+        g_free (log_path);
+
         if (! ret) {
                 g_warning ("Could not start command '%s': %s",
                            welcome_session->priv->command,
@@ -740,32 +786,13 @@ gdm_welcome_session_start (GdmWelcomeSession *welcome_session)
         return res;
 }
 
-static int
-wait_on_child (int pid)
-{
-        int status;
-
- wait_again:
-        if (waitpid (pid, &status, 0) < 0) {
-                if (errno == EINTR) {
-                        goto wait_again;
-                } else if (errno == ECHILD) {
-                        ; /* do nothing, child already reaped */
-                } else {
-                        g_debug ("GdmWelcomeSession: waitpid () should not fail");
-                }
-        }
-
-        return status;
-}
-
 static void
 welcome_session_died (GdmWelcomeSession *welcome_session)
 {
         int exit_status;
 
         g_debug ("GdmWelcomeSession: Waiting on process %d", welcome_session->priv->pid);
-        exit_status = wait_on_child (welcome_session->priv->pid);
+        exit_status = gdm_wait_on_pid (welcome_session->priv->pid);
 
         if (WIFEXITED (exit_status) && (WEXITSTATUS (exit_status) != 0)) {
                 g_debug ("GdmWelcomeSession: Wait on child process failed");
@@ -782,6 +809,7 @@ welcome_session_died (GdmWelcomeSession *welcome_session)
 gboolean
 gdm_welcome_session_stop (GdmWelcomeSession *welcome_session)
 {
+        int res;
 
         if (welcome_session->priv->pid <= 1) {
                 return TRUE;
@@ -795,8 +823,12 @@ gdm_welcome_session_stop (GdmWelcomeSession *welcome_session)
 
         g_debug ("GdmWelcomeSession: Stopping welcome_session");
 
-        gdm_signal_pid (-1 * welcome_session->priv->pid, SIGTERM);
-        welcome_session_died (welcome_session);
+        res = gdm_signal_pid (-1 * welcome_session->priv->pid, SIGTERM);
+        if (res < 0) {
+                g_warning ("Unable to kill welcome session process");
+        } else {
+                welcome_session_died (welcome_session);
+        }
 
         if (welcome_session->priv->ckc != NULL) {
                 close_welcome_session (welcome_session);
