@@ -25,6 +25,9 @@
 #include <unistd.h>
 #include <string.h>
 
+#include <dbus/dbus-glib.h>
+#include <dbus/dbus-glib-lowlevel.h>
+
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <glib-object.h>
@@ -37,14 +40,22 @@
 
 #include "gdm-session-manager.h"
 #include "gdm-session-client.h"
+#include "gdm-profile.h"
 
 #define GDM_GREETER_SESSION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GDM_TYPE_GREETER_SESSION, GdmGreeterSessionPrivate))
 
-#define KEY_GDM_A11Y_DIR            "/apps/gdm/simple-greeter/accessibility"
+#define GSD_DBUS_NAME      "org.gnome.SettingsDaemon"
+#define GSD_DBUS_PATH      "/org/gnome/SettingsDaemon"
+#define GSD_DBUS_INTERFACE "org.gnome.SettingsDaemon"
+
+#define KEY_GDM_DIR                 "/apps/gdm/simple-greeter"
+#define KEY_GDM_A11Y_DIR             KEY_GDM_DIR "/accessibility"
 #define KEY_SCREEN_KEYBOARD_ENABLED  KEY_GDM_A11Y_DIR "/screen_keyboard_enabled"
 #define KEY_SCREEN_MAGNIFIER_ENABLED KEY_GDM_A11Y_DIR "/screen_magnifier_enabled"
 #define KEY_SCREEN_READER_ENABLED    KEY_GDM_A11Y_DIR "/screen_reader_enabled"
 
+#define KEY_WM_USE_COMPIZ            KEY_GDM_DIR "/wm_use_compiz"
+
 struct GdmGreeterSessionPrivate
 {
         GdmGreeterClient      *client;
@@ -105,7 +116,7 @@ on_reset (GdmGreeterClient  *client,
 {
         g_debug ("GdmGreeterSession: Reset");
 
-        gdm_greeter_panel_hide_user_options (GDM_GREETER_PANEL (session->priv->panel));
+        gdm_greeter_panel_reset (GDM_GREETER_PANEL (session->priv->panel));
         gdm_greeter_login_window_reset (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window));
 }
 
@@ -117,6 +128,54 @@ on_selected_user_changed (GdmGreeterClient  *client,
         g_debug ("GdmGreeterSession: selected user changed: %s", text);
 }
 
+static void
+on_default_language_name_changed (GdmGreeterClient  *client,
+                                  const char        *text,
+                                  GdmGreeterSession *session)
+{
+        g_debug ("GdmGreeterSession: default language name changed: %s", text);
+        gdm_greeter_panel_set_default_language_name (GDM_GREETER_PANEL (session->priv->panel),
+                                                     text);
+}
+
+static void
+on_default_layout_name_changed (GdmGreeterClient  *client,
+                                const char        *text,
+                                GdmGreeterSession *session)
+{
+        g_debug ("GdmGreeterSession: default layout name changed: %s", text);
+        gdm_greeter_panel_set_default_layout_name (GDM_GREETER_PANEL (session->priv->panel),
+                                                     text);
+}
+
+static void
+on_default_session_name_changed (GdmGreeterClient  *client,
+                                 const char        *text,
+                                 GdmGreeterSession *session)
+{
+        g_debug ("GdmGreeterSession: default session name changed: %s", text);
+        gdm_greeter_panel_set_default_session_name (GDM_GREETER_PANEL (session->priv->panel),
+                                                    text);
+}
+
+static void
+on_timed_login_requested (GdmGreeterClient  *client,
+                          const char        *text,
+                          int                delay,
+                          GdmGreeterSession *session)
+{
+        g_debug ("GdmGreeterSession: timed login requested for user %s (in %d seconds)", text, delay);
+        gdm_greeter_login_window_request_timed_login (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window), text, delay);
+}
+
+static void
+on_user_authorized (GdmGreeterClient  *client,
+                    GdmGreeterSession *session)
+{
+        g_debug ("GdmGreeterSession: user authorized");
+        gdm_greeter_login_window_user_authorized (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window));
+}
+
 static void
 on_info_query (GdmGreeterClient  *client,
                const char        *text,
@@ -137,6 +196,15 @@ on_secret_info_query (GdmGreeterClient  *client,
         gdm_greeter_login_window_secret_info_query (GDM_GREETER_LOGIN_WINDOW (session->priv->login_window), text);
 }
 
+static void
+on_begin_auto_login (GdmGreeterLoginWindow *login_window,
+                     const char            *username,
+                     GdmGreeterSession     *session)
+{
+        gdm_greeter_client_call_begin_auto_login (session->priv->client,
+                                                  username);
+}
+
 static void
 on_begin_verification (GdmGreeterLoginWindow *login_window,
                        GdmGreeterSession     *session)
@@ -178,6 +246,14 @@ on_select_language (GdmGreeterSession     *session,
                                                  text);
 }
 
+static void
+on_select_layout (GdmGreeterSession      *session,
+                  const char             *text)
+{
+        gdm_greeter_client_call_select_layout (session->priv->client,
+                                               text);
+}
+
 static void
 on_select_user (GdmGreeterLoginWindow *login_window,
                 const char            *text,
@@ -203,11 +279,20 @@ on_disconnected (GdmGreeterLoginWindow *login_window,
         gdm_greeter_client_call_disconnect (session->priv->client);
 }
 
+static void
+on_start_session (GdmGreeterLoginWindow *login_window,
+                  GdmGreeterSession     *session)
+{
+        gdm_greeter_client_call_start_session_when_ready (session->priv->client, TRUE);
+}
+
 static void
 toggle_panel (GdmSessionManager *manager,
               gboolean           enabled,
               GdmGreeterSession *session)
 {
+        gdm_profile_start (NULL);
+
         if (enabled) {
                 session->priv->panel = gdm_greeter_panel_new ();
 
@@ -216,6 +301,11 @@ toggle_panel (GdmSessionManager *manager,
                                           G_CALLBACK (on_select_language),
                                           session);
 
+                g_signal_connect_swapped (session->priv->panel,
+                                          "layout-selected",
+                                          G_CALLBACK (on_select_layout),
+                                          session);
+
                 g_signal_connect_swapped (session->priv->panel,
                                           "session-selected",
                                           G_CALLBACK (on_select_session),
@@ -226,6 +316,8 @@ toggle_panel (GdmSessionManager *manager,
                 gtk_widget_destroy (session->priv->panel);
                 session->priv->panel = NULL;
         }
+
+        gdm_profile_end (NULL);
 }
 
 static void
@@ -233,6 +325,8 @@ toggle_login_window (GdmSessionManager *manager,
                      gboolean           enabled,
                      GdmGreeterSession *session)
 {
+        gdm_profile_start (NULL);
+
         if (enabled) {
                 gboolean is_local;
 
@@ -240,6 +334,10 @@ toggle_login_window (GdmSessionManager *manager,
                 g_debug ("GdmGreeterSession: Starting a login window local:%d", is_local);
                 session->priv->login_window = gdm_greeter_login_window_new (is_local);
 
+                g_signal_connect (session->priv->login_window,
+                                  "begin-auto-login",
+                                  G_CALLBACK (on_begin_auto_login),
+                                  session);
                 g_signal_connect (session->priv->login_window,
                                   "begin-verification",
                                   G_CALLBACK (on_begin_verification),
@@ -264,11 +362,16 @@ toggle_login_window (GdmSessionManager *manager,
                                   "disconnected",
                                   G_CALLBACK (on_disconnected),
                                   session);
+                g_signal_connect (session->priv->login_window,
+                                  "start-session",
+                                  G_CALLBACK (on_start_session),
+                                  session);
                 gtk_widget_show (session->priv->login_window);
         } else {
                 gtk_widget_destroy (session->priv->login_window);
                 session->priv->login_window = NULL;
         }
+        gdm_profile_end (NULL);
 }
 
 static gboolean
@@ -277,12 +380,15 @@ launch_compiz (GdmGreeterSession *session)
         GError  *error;
         gboolean ret;
 
+        gdm_profile_start (NULL);
         g_debug ("GdmGreeterSession: Launching compiz");
 
         ret = FALSE;
 
+        g_setenv ("LIBGL_ALWAYS_INDIRECT", "1", TRUE);
+
         error = NULL;
-        g_spawn_command_line_async ("gtk-window-decorator --replace", &error);
+        g_spawn_command_line_async ("gtk-window-decorator", &error);
         if (error != NULL) {
                 g_warning ("Error starting WM: %s", error->message);
                 g_error_free (error);
@@ -290,7 +396,7 @@ launch_compiz (GdmGreeterSession *session)
         }
 
         error = NULL;
-        g_spawn_command_line_async ("compiz --replace", &error);
+        g_spawn_command_line_async ("compiz --replace glib gconf", &error);
         if (error != NULL) {
                 g_warning ("Error starting WM: %s", error->message);
                 g_error_free (error);
@@ -302,6 +408,7 @@ launch_compiz (GdmGreeterSession *session)
         /* FIXME: should try to detect if it actually works */
 
  out:
+        gdm_profile_end (NULL);
         return ret;
 }
 
@@ -311,6 +418,8 @@ launch_metacity (GdmGreeterSession *session)
         GError  *error;
         gboolean ret;
 
+        gdm_profile_start (NULL);
+
         g_debug ("GdmGreeterSession: Launching metacity");
 
         ret = FALSE;
@@ -326,15 +435,26 @@ launch_metacity (GdmGreeterSession *session)
         ret = TRUE;
 
  out:
+        gdm_profile_end (NULL);
         return ret;
 }
 
 static void
 start_window_manager (GdmGreeterSession *session)
 {
-        if (! launch_metacity (session)) {
-                launch_compiz (session);
+        gboolean     use_compiz;
+        GConfClient *client;
+
+        client = gconf_client_get_default ();
+        use_compiz = gconf_client_get_bool (client, KEY_WM_USE_COMPIZ, NULL);
+        g_object_unref (client);
+        if (use_compiz) {
+                if (launch_compiz (session)) {
+                        return;
+                }
         }
+
+        launch_metacity (session);
 }
 
 static void
@@ -437,7 +557,7 @@ setup_at_tools (GdmGreeterSession *session)
         gdm_session_client_set_try_exec (session->priv->screen_keyboard_client,
                                          "gok");
         gdm_session_client_set_command (session->priv->screen_keyboard_client,
-                                        "gok");
+                                        "gok --login");
         enabled = gconf_client_get_bool (client, KEY_SCREEN_KEYBOARD_ENABLED, NULL);
         gdm_session_client_set_enabled (session->priv->screen_keyboard_client,
                                         enabled);
@@ -449,7 +569,7 @@ setup_at_tools (GdmGreeterSession *session)
         gdm_session_client_set_try_exec (session->priv->screen_reader_client,
                                          "orca");
         gdm_session_client_set_command (session->priv->screen_reader_client,
-                                        "orca -n");
+                                        "orca -n -d main-window -d magnifier");
         enabled = gconf_client_get_bool (client, KEY_SCREEN_READER_ENABLED, NULL);
         gdm_session_client_set_enabled (session->priv->screen_reader_client,
                                         enabled);
@@ -479,6 +599,98 @@ setup_at_tools (GdmGreeterSession *session)
         g_object_unref (client);
 }
 
+static gboolean
+send_dbus_string_method (DBusConnection *connection,
+                         const char     *method,
+                         const char     *payload)
+{
+        DBusError       error;
+        DBusMessage    *message;
+        DBusMessage    *reply;
+        DBusMessageIter iter;
+        const char     *str;
+
+        if (payload != NULL) {
+                str = payload;
+        } else {
+                str = "";
+        }
+
+        g_debug ("GdmGreeterSession: Calling %s", method);
+        message = dbus_message_new_method_call (GSD_DBUS_NAME,
+                                                GSD_DBUS_PATH,
+                                                GSD_DBUS_INTERFACE,
+                                                method);
+        if (message == NULL) {
+                g_warning ("Couldn't allocate the D-Bus message");
+                return FALSE;
+        }
+
+        dbus_message_iter_init_append (message, &iter);
+        dbus_message_iter_append_basic (&iter,
+                                        DBUS_TYPE_STRING,
+                                        &str);
+
+        dbus_error_init (&error);
+        reply = dbus_connection_send_with_reply_and_block (connection,
+                                                           message,
+                                                           -1,
+                                                           &error);
+
+        dbus_message_unref (message);
+
+        if (dbus_error_is_set (&error)) {
+                g_warning ("%s %s raised: %s\n",
+                           method,
+                           error.name,
+                           error.message);
+                return FALSE;
+        }
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
+        dbus_connection_flush (connection);
+
+        return TRUE;
+}
+
+static gboolean
+activate_settings_daemon (GdmGreeterSession *session)
+{
+        gboolean         ret;
+        gboolean         res;
+        DBusError        local_error;
+        DBusConnection  *connection;
+
+        gdm_profile_start (NULL);
+
+        ret = FALSE;
+
+        g_debug ("GdmGreeterLoginWindow: activating settings daemon");
+
+        dbus_error_init (&local_error);
+        connection = dbus_bus_get (DBUS_BUS_SESSION, &local_error);
+        if (connection == NULL) {
+                g_debug ("Failed to connect to the D-Bus daemon: %s", local_error.message);
+                dbus_error_free (&local_error);
+                goto out;
+        }
+
+        res = send_dbus_string_method (connection,
+                                       "StartWithSettingsPrefix",
+                                       "/apps/gdm/simple-greeter/settings-manager-plugins");
+        if (! res) {
+                g_warning ("Couldn't start settings daemon");
+                goto out;
+        }
+        ret = TRUE;
+        g_debug ("GdmGreeterLoginWindow: settings daemon started");
+ out:
+        gdm_profile_end (NULL);
+
+        return ret;
+}
+
 static gboolean
 start_settings_daemon (GdmGreeterSession *session)
 {
@@ -523,10 +735,14 @@ gdm_greeter_session_start (GdmGreeterSession *session,
 
         g_return_val_if_fail (GDM_IS_GREETER_SESSION (session), FALSE);
 
+        gdm_profile_start (NULL);
+
         res = gdm_greeter_client_start (session->priv->client, error);
 
         gdm_session_manager_set_level (session->priv->manager, GDM_SESSION_LEVEL_LOGIN_WINDOW);
 
+        gdm_profile_end (NULL);
+
         return res;
 }
 
@@ -543,10 +759,6 @@ gdm_greeter_session_set_property (GObject        *object,
                                   const GValue   *value,
                                   GParamSpec     *pspec)
 {
-        GdmGreeterSession *self;
-
-        self = GDM_GREETER_SESSION (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -560,10 +772,6 @@ gdm_greeter_session_get_property (GObject        *object,
                                   GValue         *value,
                                   GParamSpec     *pspec)
 {
-        GdmGreeterSession *self;
-
-        self = GDM_GREETER_SESSION (object);
-
         switch (prop_id) {
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -577,9 +785,6 @@ gdm_greeter_session_constructor (GType                  type,
                                  GObjectConstructParam *construct_properties)
 {
         GdmGreeterSession      *greeter_session;
-        GdmGreeterSessionClass *klass;
-
-        klass = GDM_GREETER_SESSION_CLASS (g_type_class_peek (GDM_TYPE_GREETER_SESSION));
 
         greeter_session = GDM_GREETER_SESSION (G_OBJECT_CLASS (gdm_greeter_session_parent_class)->constructor (type,
                                                                                                                n_construct_properties,
@@ -591,10 +796,6 @@ gdm_greeter_session_constructor (GType                  type,
 static void
 gdm_greeter_session_dispose (GObject *object)
 {
-        GdmGreeterSession *greeter_session;
-
-        greeter_session = GDM_GREETER_SESSION (object);
-
         g_debug ("GdmGreeterSession: Disposing greeter_session");
 
         G_OBJECT_CLASS (gdm_greeter_session_parent_class)->dispose (object);
@@ -649,6 +850,7 @@ gdm_greeter_session_event_handler(GdkEvent          *event,
 static void
 gdm_greeter_session_init (GdmGreeterSession *session)
 {
+        gdm_profile_start (NULL);
 
         session->priv = GDM_GREETER_SESSION_GET_PRIVATE (session);
 
@@ -697,6 +899,26 @@ gdm_greeter_session_init (GdmGreeterSession *session)
                           "selected-user-changed",
                           G_CALLBACK (on_selected_user_changed),
                           session);
+        g_signal_connect (session->priv->client,
+                          "default-language-name-changed",
+                          G_CALLBACK (on_default_language_name_changed),
+                          session);
+        g_signal_connect (session->priv->client,
+                          "default-layout-name-changed",
+                          G_CALLBACK (on_default_layout_name_changed),
+                          session);
+        g_signal_connect (session->priv->client,
+                          "default-session-name-changed",
+                          G_CALLBACK (on_default_session_name_changed),
+                          session);
+        g_signal_connect (session->priv->client,
+                          "timed-login-requested",
+                          G_CALLBACK (on_timed_login_requested),
+                          session);
+        g_signal_connect (session->priv->client,
+                          "user-authorized",
+                          G_CALLBACK (on_user_authorized),
+                          session);
 
         /* We want to listen for panel mnemonics even if the
          * login window is focused, so we intercept them here.
@@ -706,6 +928,8 @@ gdm_greeter_session_init (GdmGreeterSession *session)
 
         /* FIXME: we should really do this in settings daemon */
         setup_at_tools (session);
+
+        gdm_profile_end (NULL);
 }
 
 static void
